//
// Syd: rock-solid unikernel
// src/globset.rs: GlobSet matcher
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
// Based in part upon ripgrep which is:
//   Copyright (c) 2015 Andrew Gallant
//   SPDX-License-Identifier: MIT
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Last sync with globset:
// Version 0.4.14
// Commit:c9ebcbd8abe48c8336fb4826df7e9b6fb179de03

/*!
The globset crate provides cross platform single glob and glob set matching.

Glob set matching is the process of matching one or more glob patterns against
a single candidate path simultaneously, and returning all of the globs that
matched. For example, given this set of globs:

* `*.rs`
* `src/lib.rs`
* `src/**/foo.rs`

and a path `src/bar/baz/foo.rs`, then the set would report the first and third
globs as matching.

# Example: one glob

This example shows how to match a single glob against a single file path.

```
use std::path::Path;
use syd::globset::Glob;

let glob = Glob::new("*.rs")?.compile_matcher();

assert!(glob.is_match(Path::new("foo.rs")));
assert!(glob.is_match(Path::new("foo/bar.rs")));
assert!(!glob.is_match(Path::new("Cargo.toml")));
# Ok::<(), Box<dyn std::error::Error>>(())
```

# Example: configuring a glob matcher

This example shows how to use a `GlobBuilder` to configure aspects of match
semantics. In this example, we prevent wildcards from matching path separators.

```
use std::path::Path;
use syd::globset::GlobBuilder;

let glob = GlobBuilder::new("*.rs")
    .literal_separator(true).build()?.compile_matcher();

assert!(glob.is_match(Path::new("foo.rs")));
assert!(!glob.is_match(Path::new("foo/bar.rs"))); // no longer matches
assert!(!glob.is_match(Path::new("Cargo.toml")));
# Ok::<(), Box<dyn std::error::Error>>(())
```

# Example: match multiple globs at once

This example shows how to match multiple glob patterns at once.

```
use std::path::Path;
use syd::globset::{Glob, GlobSetBuilder};

let mut builder = GlobSetBuilder::new();
// A GlobBuilder can be used to configure each glob's match semantics
// independently.
builder.add(Glob::new("*.rs")?);
builder.add(Glob::new("src/lib.rs")?);
builder.add(Glob::new("src/**/foo.rs")?);
let set = builder.build()?;

assert_eq!(set.matches(Path::new("src/bar/baz/foo.rs")), Some(2));
# Ok::<(), Box<dyn std::error::Error>>(())
```

# Syntax

Standard Unix-style glob syntax is supported:

* `?` matches any single character. (If the `literal_separator` option is
  enabled, then `?` can never match a path separator.)
* `*` matches zero or more characters. (If the `literal_separator` option is
  enabled, then `*` can never match a path separator.)
* `**` recursively matches directories but are only legal in three situations.
  First, if the glob starts with <code>\*\*&#x2F;</code>, then it matches
  all directories. For example, <code>\*\*&#x2F;foo</code> matches `foo`
  and `bar/foo` but not `foo/bar`. Secondly, if the glob ends with
  <code>&#x2F;\*\*</code>, then it matches all sub-entries. For example,
  <code>foo&#x2F;\*\*</code> matches `foo/a` and `foo/a/b`, but not `foo`.
  Thirdly, if the glob contains <code>&#x2F;\*\*&#x2F;</code> anywhere within
  the pattern, then it matches zero or more directories. Using `**` anywhere
  else is illegal (N.B. the glob `**` is allowed and means "match everything").
* `{a,b}` matches `a` or `b` where `a` and `b` are arbitrary glob patterns.
  (N.B. Nesting `{...}` is not currently allowed.)
* `[ab]` matches `a` or `b` where `a` and `b` are characters. Use
  `[!ab]` to match any character except for `a` and `b`.
* Metacharacters such as `*` and `?` can be escaped with character class
  notation. e.g., `[*]` matches `*`.
* When backslash escapes are enabled, a backslash (`\`) will escape all meta
  characters in a glob. If it precedes a non-meta character, then the slash is
  ignored. A `\\` will match a literal `\\`. Note that this mode is only
  enabled on Unix platforms by default, but can be enabled on any platform
  via the `backslash_escape` setting on `Glob`.

A `GlobBuilder` can be used to prevent wildcards from matching path separators,
or to enable case insensitive matching.
*/

#![deny(missing_docs)]

use std::{collections::VecDeque, os::unix::ffi::OsStrExt, path::Path};

use regex_automata::{
    dfa::{dense::DFA, Automaton, StartKind},
    nfa::thompson,
    util::syntax,
    Anchored, Input, MatchKind,
};

pub use crate::glob::{Glob, GlobBuilder, GlobMatcher};

/// Represents an error that can occur when parsing a glob pattern.
#[derive(Clone, Debug, Eq, PartialEq)]
pub struct Error {
    /// The original glob provided by the caller.
    pub glob: Option<String>,
    /// The kind of error.
    pub kind: ErrorKind,
}

/// The kind of error that can occur when parsing a glob pattern.
#[derive(Clone, Debug, Eq, PartialEq)]
pub enum ErrorKind {
    /// **DEPRECATED**.
    ///
    /// This error used to occur for consistency with git's glob specification,
    /// but the specification now accepts all uses of `**`. When `**` does not
    /// appear adjacent to a path separator or at the beginning/end of a glob,
    /// it is now treated as two consecutive `*` patterns. As such, this error
    /// is no longer used.
    InvalidRecursive,
    /// Occurs when a character class (e.g., `[abc]`) is not closed.
    UnclosedClass,
    /// Occurs when a range in a character (e.g., `[a-z]`) is invalid. For
    /// example, if the range starts with a lexicographically larger character
    /// than it ends with.
    InvalidRange(char, char),
    /// Occurs when a `}` is found without a matching `{`.
    UnopenedAlternates,
    /// Occurs when a `{` is found without a matching `}`.
    UnclosedAlternates,
    /// Occurs when an alternating group is nested inside another alternating
    /// group, e.g., `{{a,b},{c,d}}`.
    NestedAlternates,
    /// Occurs when an unescaped '\' is found at the end of a glob.
    DanglingEscape,
    /// An error associated with parsing or compiling a regex.
    Regex(String),
    /// Hints that destructuring should not be exhaustive.
    ///
    /// This enum may grow additional variants, so this makes sure clients
    /// don't count on exhaustive matching. (Otherwise, adding a new variant
    /// could break existing code.)
    #[doc(hidden)]
    __Nonexhaustive,
}

impl std::error::Error for Error {
    fn description(&self) -> &str {
        self.kind.description()
    }
}

impl Error {
    /// Return the glob that caused this error, if one exists.
    pub fn glob(&self) -> Option<&str> {
        self.glob.as_deref()
    }

    /// Return the kind of this error.
    pub fn kind(&self) -> &ErrorKind {
        &self.kind
    }
}

impl ErrorKind {
    fn description(&self) -> &str {
        match *self {
            ErrorKind::InvalidRecursive => "invalid use of **; must be one path component",
            ErrorKind::UnclosedClass => "unclosed character class; missing ']'",
            ErrorKind::InvalidRange(_, _) => "invalid character range",
            ErrorKind::UnopenedAlternates => {
                "unopened alternate group; missing '{' \
                (maybe escape '}' with '[}]'?)"
            }
            ErrorKind::UnclosedAlternates => {
                "unclosed alternate group; missing '}' \
                (maybe escape '{' with '[{]'?)"
            }
            ErrorKind::NestedAlternates => "nested alternate groups are not allowed",
            ErrorKind::DanglingEscape => "dangling '\\'",
            ErrorKind::Regex(ref err) => err,
            ErrorKind::__Nonexhaustive => unreachable!(),
        }
    }
}

impl std::fmt::Display for Error {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self.glob {
            None => self.kind.fmt(f),
            Some(ref glob) => {
                write!(f, "error parsing glob '{}': {}", glob, self.kind)
            }
        }
    }
}

impl std::fmt::Display for ErrorKind {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match *self {
            ErrorKind::InvalidRecursive
            | ErrorKind::UnclosedClass
            | ErrorKind::UnopenedAlternates
            | ErrorKind::UnclosedAlternates
            | ErrorKind::NestedAlternates
            | ErrorKind::DanglingEscape
            | ErrorKind::Regex(_) => write!(f, "{}", self.description()),
            ErrorKind::InvalidRange(s, e) => {
                write!(f, "invalid range; '{}' > '{}'", s, e)
            }
            ErrorKind::__Nonexhaustive => unreachable!(),
        }
    }
}

#[allow(clippy::disallowed_methods)]
pub(crate) fn new_regex(pat: &str) -> Result<DFA<Vec<u32>>, Error> {
    DFA::builder()
        .configure(
            DFA::config()
                .accelerate(true)
                .match_kind(MatchKind::LeftmostFirst)
                .start_kind(StartKind::Anchored),
        )
        .syntax(syntax::Config::new().unicode(false).utf8(false))
        .thompson(thompson::Config::new().shrink(true).utf8(false))
        .build(pat)
        .map_err(|err| Error {
            glob: Some(pat.to_string()),
            kind: ErrorKind::Regex(err.to_string()),
        })
}

#[allow(clippy::disallowed_methods)]
fn new_regex_set(pats: &[String]) -> Result<DFA<Vec<u32>>, Error> {
    DFA::builder()
        .configure(
            DFA::config()
                .accelerate(true)
                .match_kind(MatchKind::LeftmostFirst)
                .start_kind(StartKind::Anchored),
        )
        .syntax(syntax::Config::new().unicode(false).utf8(false))
        .thompson(thompson::Config::new().shrink(true).utf8(false))
        .build_many(pats)
        .map_err(|err| Error {
            glob: None,
            kind: ErrorKind::Regex(err.to_string()),
        })
}

#[derive(Clone, Debug)]
struct GlobSetMatchStrategy(RegexSetStrategy);

impl GlobSetMatchStrategy {
    fn is_match(&self, path: &Path) -> bool {
        self.0.is_match(path)
    }

    fn matches(&self, path: &Path) -> Option<usize> {
        self.0.matches(path)
    }
}

#[derive(Clone, Debug)]
struct MultiStrategyBuilder(VecDeque<String>);

impl MultiStrategyBuilder {
    fn new() -> Self {
        Self(VecDeque::with_capacity(128))
    }

    fn add(&mut self, literal: String) {
        self.0.push_front(literal);
    }

    fn regex_set(mut self) -> Result<RegexSetStrategy, Error> {
        self.0.make_contiguous();
        Ok(RegexSetStrategy(new_regex_set(self.0.as_slices().0)?))
    }
}

/// GlobSet represents a group of globs that can be matched together in a
/// single pass.
#[derive(Clone, Debug)]
pub struct GlobSet(Option<GlobSetMatchStrategy>);

impl GlobSet {
    /// Create a new [`GlobSetBuilder`]. A `GlobSetBuilder` can be used to add
    /// new patterns. Once all patterns have been added, `build` should be
    /// called to produce a `GlobSet`, which can then be used for matching.
    #[inline]
    pub fn builder() -> GlobSetBuilder {
        GlobSetBuilder::new()
    }

    /// Create an empty `GlobSet`. An empty set matches nothing.
    #[inline]
    pub fn empty() -> Self {
        Self(None)
    }

    /// Returns true if this set is empty, and therefore matches nothing.
    #[inline]
    pub fn is_empty(&self) -> bool {
        self.0.is_none()
    }

    /// Returns true if any glob in this set matches the path given.
    pub fn is_match(&self, path: &Path) -> bool {
        self.0.as_ref().map_or(false, |strat| strat.is_match(path))
    }

    /// Returns the last matching index or None.
    pub fn matches(&self, path: &Path) -> Option<usize> {
        self.0.as_ref().and_then(|strat| strat.matches(path))
    }

    fn new(pats: &[Glob]) -> Result<GlobSet, Error> {
        if pats.is_empty() {
            return Ok(GlobSet(None));
        }
        let mut regexes = MultiStrategyBuilder::new();
        for p in pats {
            regexes.add(p.regex().to_owned());
        }
        Ok(GlobSet(Some(GlobSetMatchStrategy(regexes.regex_set()?))))
    }
}

impl Default for GlobSet {
    /// Create a default empty GlobSet.
    fn default() -> Self {
        GlobSet::empty()
    }
}

/// GlobSetBuilder builds a group of patterns that can be used to
/// simultaneously match a file path.
#[derive(Clone, Debug)]
pub struct GlobSetBuilder {
    pats: Vec<Glob>,
}

impl GlobSetBuilder {
    /// Create a new `GlobSetBuilder`. A `GlobSetBuilder` can be used to add new
    /// patterns. Once all patterns have been added, `build` should be called
    /// to produce a [`GlobSet`], which can then be used for matching.
    pub fn new() -> GlobSetBuilder {
        GlobSetBuilder { pats: vec![] }
    }

    /// Builds a new matcher from all of the glob patterns added so far.
    ///
    /// Once a matcher is built, no new patterns can be added to it.
    pub fn build(&self) -> Result<GlobSet, Error> {
        GlobSet::new(&self.pats)
    }

    /// Add a new pattern to this set.
    pub fn add(&mut self, pat: Glob) -> &mut GlobSetBuilder {
        self.pats.push(pat);
        self
    }
}

impl Default for GlobSetBuilder {
    fn default() -> Self {
        Self::new()
    }
}

#[derive(Clone, Debug)]
struct RegexSetStrategy(DFA<Vec<u32>>);

impl RegexSetStrategy {
    fn is_match(&self, path: &Path) -> bool {
        self.matches(path).is_some()
    }

    fn matches(&self, path: &Path) -> Option<usize> {
        let input = Input::new(path.as_os_str().as_bytes()).anchored(Anchored::Yes);

        // SAFETY: MatchError does not mean there was no match!
        #[allow(clippy::arithmetic_side_effects)]
        #[allow(clippy::disallowed_methods)]
        self.0
            .try_search_fwd(&input)
            .expect("try_search_fwd")
            .map(|half_match| self.0.pattern_len() - (half_match.pattern().as_usize() + 1))
    }
}

/// Escape meta-characters within the given glob pattern.
///
/// The escaping works by surrounding meta-characters with brackets. For
/// example, `*` becomes `[*]`.
#[inline(always)]
pub fn escape(s: &str) -> String {
    let mut escaped = String::with_capacity(s.len());
    for c in s.chars() {
        match c {
            // Note that ! does not need escaping because it is only
            // special inside brackets.
            '?' | '*' | '[' | ']' => {
                escaped.push('[');
                escaped.push(c);
                escaped.push(']');
            }
            _ => escaped.push(c),
        }
    }
    escaped
}

#[cfg(test)]
mod tests {
    use std::path::Path;

    use super::{GlobSet, GlobSetBuilder};
    use crate::glob::Glob;

    #[test]
    fn set_works() {
        let mut builder = GlobSetBuilder::new();
        builder.add(Glob::new("src/**/*.rs").unwrap());
        builder.add(Glob::new("*.c").unwrap());
        builder.add(Glob::new("src/lib.rs").unwrap());
        let set = builder.build().unwrap();

        assert!(set.is_match(Path::new("foo.c")));
        assert!(set.is_match(Path::new("src/foo.c")));
        assert!(!set.is_match(Path::new("foo.rs")));
        assert!(!set.is_match(Path::new("tests/foo.rs")));
        assert!(set.is_match(Path::new("src/foo.rs")));
        assert!(set.is_match(Path::new("src/grep/src/main.rs")));

        assert_eq!(Some(2), set.matches(Path::new("src/lib.rs")));
    }

    #[test]
    fn empty_set_works() {
        let set = GlobSetBuilder::new().build().unwrap();
        assert!(!set.is_match(Path::new("")));
        assert!(!set.is_match(Path::new("a")));
    }

    #[test]
    fn default_set_is_empty_works() {
        let set: GlobSet = Default::default();
        assert!(!set.is_match(Path::new("")));
        assert!(!set.is_match(Path::new("a")));
    }

    #[test]
    fn escape() {
        use super::escape;
        assert_eq!("foo", escape("foo"));
        assert_eq!("foo[*]", escape("foo*"));
        assert_eq!("[[][]]", escape("[]"));
        assert_eq!("[*][?]", escape("*?"));
        assert_eq!("src/[*][*]/[*].rs", escape("src/**/*.rs"));
        assert_eq!("bar[[]ab[]]baz", escape("bar[ab]baz"));
        assert_eq!("bar[[]!![]]!baz", escape("bar[!!]!baz"));
    }

    // This tests that regex matching doesn't "remember" the results of
    // previous searches. That is, if any memory is reused from a previous
    // search, then it should be cleared first.
    #[test]
    fn set_does_not_remember() {
        let mut builder = GlobSetBuilder::new();
        builder.add(Glob::new("*foo*").unwrap());
        builder.add(Glob::new("*bar*").unwrap());
        builder.add(Glob::new("*quux*").unwrap());
        let set = builder.build().unwrap();

        assert_eq!(Some(2), set.matches(Path::new("ZfooZquuxZ")));
        assert_eq!(None, set.matches(Path::new("nada")));
    }
}
