//
// Syd: rock-solid unikernel
// src/pool.rs: Self growing / shrinking `ThreadPool` implementation
//
// Copyright (c) 2024 Ali Polatel <alip@chesswob.org>
// Based in part upon rusty_pool which is:
//     Copyright (c) Robin Friedli <robinfriedli@icloud.com>
//     SPDX-License-Identifier: Apache-2.0
//
// SPDX-License-Identifier: GPL-3.0-or-later

// Last sync with rusty_pool:
// Version 0.7.0
// Commit:d56805869ba3cbe47021d5660bbaf19ac5ec4bfb

use std::{
    option::Option,
    sync::{
        atomic::{AtomicU32, Ordering},
        Arc,
    },
    thread,
    time::Duration,
};

use once_cell::sync::Lazy;

use crate::config::EMU_STACK_SIZE;

#[allow(clippy::cast_possible_truncation)]
static NPROC: Lazy<u16> = Lazy::new(|| num_cpus::get() as u16);

/// The absolute maximum number of workers. This corresponds to the
/// maximum value that can be stored within half the bits of u32, as two
/// counters (total workers and idle workers) are stored in one
/// AtomicU32.
pub(crate) const MAX_SIZE: u16 = u16::MAX;

type Job = Box<dyn FnOnce() + Send + 'static>;

/// Trait to implement for all items that may be executed by the
/// `ThreadPool`.
pub(crate) trait Task<R: Send>: Send {
    /// Execute this task and return its result.
    fn run(self) -> R;

    /// Transform this `Task` into a heap allocated `FnOnce` if
    /// possible.
    ///
    /// Used by
    /// [`ThreadPool::execute`](struct.ThreadPool.html#method.execute)
    /// to turn this `Task` into a `Job` directly without having to
    /// create an additional `Job` that calls this `Task`.
    fn into_fn(self) -> Option<Box<dyn FnOnce() -> R + Send + 'static>>;

    /// Return `true` if calling [`Task::into_fn`] on this `Task` returns `Some`.
    fn is_fn(&self) -> bool;
}

/// Implement the `Task` trait for any FnOnce closure that returns a
/// thread-safe result.
impl<R, F> Task<R> for F
where
    R: Send,
    F: FnOnce() -> R + Send + 'static,
{
    fn run(self) -> R {
        self()
    }

    fn into_fn(self) -> Option<Box<dyn FnOnce() -> R + Send + 'static>> {
        Some(Box::new(self))
    }

    fn is_fn(&self) -> bool {
        true
    }
}

/// Self growing / shrinking `ThreadPool` implementation based on
/// crossbeam's multi-producer multi-consumer channels that enables
/// awaiting the result of a task and offers async support.
///
/// This `ThreadPool` has two different pool sizes; a core pool size
/// filled with threads that live for as long as the channel and a max
/// pool size which describes the maximum amount of worker threads that
/// may live at the same time. Those additional non-core threads have a
/// specific keep_alive time described when creating the `ThreadPool`
/// that defines how long such threads may be idle for without receiving
/// any work before giving up and terminating their work loop.
///
/// This `ThreadPool` does not spawn any threads until a task is
/// submitted to it. Then it will create a new thread for each task
/// until the core pool size is full. After that a new thread will only
/// be created upon an `execute()` call if the current pool is lower
/// than the max pool size and there are no idle threads.
///
/// Functions like `evaluate()` and `complete()` return a `JoinHandle`
/// that may be used to await the result of a submitted task or future.
/// JoinHandles may be sent to the thread pool to create a task that
/// blocks a worker thread until it receives the result of the other
/// task and then operates on the result. If the task panics the
/// `JoinHandle` receives a cancellation error. This is implemented
/// using a futures oneshot channel to communicate with the worker
/// thread.
///
/// When creating a new worker this `ThreadPool` tries to increment the
/// worker count using a compare-and-swap mechanism, if the increment
/// fails because the total worker count has been incremented to the
/// specified limit (the core_size when trying to create a core thread,
/// else MAX_SIZE) by another thread, the pool tries to create a
/// non-core worker instead (if previously trying to create a core
/// worker and no idle worker exists) or sends the task to the channel
/// instead. Panicking workers are always cloned and replaced.
///
/// Locks are only used for the join functions to lock the `Condvar`,
/// apart from that this `ThreadPool` implementation fully relies on
/// crossbeam and atomic operations. This `ThreadPool` decides whether
/// it is currently idle (and should fast-return join attempts) by
/// comparing the total worker count to the idle worker count, which are
/// two values stored in one `AtomicU32` (both half the size of usize)
/// making sure that if both are updated they may be updated in a single
/// atomic operation.
///
/// The thread pool and its crossbeam channel can be destroyed by using
/// the shutdown function, however that does not stop tasks that are
/// already running but will terminate the thread the next time it will
/// try to fetch work from the channel. The channel is only destroyed
/// once all clones of the `ThreadPool` have been shut down / dropped.
#[derive(Clone)]
pub(crate) struct ThreadPool {
    core_size: u16,
    keep_alive: Duration,
    channel_data: Arc<ChannelData>,
    worker_data: Arc<WorkerData>,
}

impl ThreadPool {
    /// Construct a new `ThreadPool` with the specified core pool size,
    /// max pool size and keep_alive time for non-core threads. This
    /// function does not spawn any threads.
    ///
    /// `core_size` specifies the amount of threads to keep alive for as
    /// long as the `ThreadPool` exists and its channel remains
    /// connected.
    ///
    /// `keep_alive` specifies the duration for which to keep non-core
    /// pool worker threads alive while they do not receive any work.
    pub(crate) fn new(core_size: u16, keep_alive: Duration) -> Self {
        let (sender, receiver) = crossbeam_channel::unbounded();

        let worker_data = WorkerData::default();
        let channel_data = ChannelData { sender, receiver };

        Self {
            core_size,
            keep_alive,
            channel_data: Arc::new(channel_data),
            worker_data: Arc::new(worker_data),
        }
    }

    /// Send a new task to the worker threads. This function is
    /// responsible for sending the message through the channel and
    /// creating new workers if needed. If the current worker count is
    /// lower than the core pool size this function will always create a
    /// new worker. If the current worker count is equal to or greater
    /// than the core pool size this function only creates a new worker
    /// if the worker count is below the max pool size and there are no
    /// idle threads.
    ///
    /// When attempting to increment the total worker count before
    /// creating a worker fails due to the counter reaching the provided
    /// limit (core_size when attempting to create core thread, else
    /// MAX_SIZE) after being incremented by another thread, the pool
    /// tries to create a non-core worker instead (if previously trying
    /// to create a core worker and no idle worker exists) or sends the
    /// task to the channel instead. If incrementing the counter
    /// succeeded, either because the current value of the counter
    /// matched the expected value or because the last observed value
    /// was still below the limit, the worker starts with the provided
    /// task as initial task and spawns its thread.
    ///
    /// # Panics
    ///
    /// This function might panic if `try_execute` returns an error when
    /// the crossbeam channel has been closed unexpectedly.
    /// This should never occur under normal circumstances using safe
    /// code, as shutting down the `ThreadPool` consumes ownership and
    /// the crossbeam channel is never dropped unless dropping the
    /// `ThreadPool`.
    pub(crate) fn execute<T: Task<()> + 'static>(&self, task: T) {
        if self.try_execute(task).is_err() {
            panic!("the channel of the thread pool has been closed");
        }
    }

    /// Send a new task to the worker threads. This function is
    /// responsible for sending the message through the channel and
    /// creating new workers if needed. If the current worker count is
    /// lower than the core pool size this function will always create a
    /// new worker. If the current worker count is equal to or greater
    /// than the core pool size this function only creates a new worker
    /// if the worker count is below the max pool size and there are no
    /// idle threads.
    ///
    /// When attempting to increment the total worker count before
    /// creating a worker fails due to the counter reaching the provided
    /// limit (core_size when attempting to create core thread, else
    /// MAX_SIZE) after being incremented by another thread, the pool
    /// tries to create a non-core worker instead (if previously trying
    /// to create a core worker and no idle worker exists) or sends the
    /// task to the channel instead. If incrementing the counter
    /// succeeded, either because the current value of the counter
    /// matched the expected value or because the last observed value
    /// was still below the limit, the worker starts with the provided
    /// task as initial task and spawns its thread.
    ///
    /// # Errors
    ///
    /// This function might return `crossbeam_channel::SendError` if the
    /// sender was dropped unexpectedly.
    pub(crate) fn try_execute<T: Task<()> + 'static>(
        &self,
        task: T,
    ) -> Result<(), crossbeam_channel::SendError<Job>> {
        if task.is_fn() {
            #[allow(clippy::disallowed_methods)]
            self.try_execute_task(
                task.into_fn()
                    .expect("Task::into_fn returned None despite is_fn returning true"),
            )
        } else {
            self.try_execute_task(Box::new(move || {
                task.run();
            }))
        }
    }

    #[inline]
    fn try_execute_task(&self, task: Job) -> Result<(), crossbeam_channel::SendError<Job>> {
        // Create a new worker either if the current worker count is
        // lower than the core pool size or if there are no idle threads
        // and the current worker count is lower than the max pool size.
        let worker_count_val = self.worker_data.0.load(Ordering::Relaxed);
        let (curr_worker_count, idle_worker_count) = WorkerData::split(worker_count_val);

        if idle_worker_count > 0 {
            // Idle workers are available, send task!
            return self.send_task_to_channel(task);
        }

        let keep_alive = if curr_worker_count < self.core_size {
            // Create a new core worker if current pool size is below
            // core size and the pool has been observed to be busy (no
            // idle workers) during the invocation of this function.
            None
        } else if curr_worker_count < MAX_SIZE {
            // Create a new helper worker if the current worker count is
            // below the MAX_SIZE and the pool has been observed to be
            // busy (no idle workers) during the invocation of this
            // function.
            Some(self.keep_alive)
        } else {
            // No need to spawn a new worker,
            // send the task to the channel.
            return self.send_task_to_channel(task);
        };

        // Try to spawn a new worker.
        // TODO: Handle errors.
        let _ = Worker::new(
            self.channel_data.receiver.clone(),
            Arc::clone(&self.worker_data),
            keep_alive,
        )
        .try_start();

        // Finally, send the task to the channel.
        self.send_task_to_channel(task)
    }

    #[inline]
    fn send_task_to_channel(&self, task: Job) -> Result<(), crossbeam_channel::SendError<Job>> {
        self.channel_data.sender.send(task)?;

        Ok(())
    }
}

/// A helper struct to aid creating a new `ThreadPool` using default values where no value was
/// explicitly specified.
#[derive(Default)]
pub(crate) struct Builder {
    core_size: Option<u16>,
    keep_alive: Option<Duration>,
}

impl Builder {
    /// Create a new `Builder`.
    pub(crate) fn new() -> Builder {
        Builder::default()
    }

    /// Specify the core pool size for the `ThreadPool`. The core pool
    /// size is the number of threads that stay alive for the entire
    /// lifetime of the `ThreadPool` or, to be more precise, its
    /// channel. These threads are spawned if a task is submitted to the
    /// `ThreadPool` and the current worker count is below the core pool
    /// size.
    pub(crate) fn core_size(mut self, size: u16) -> Builder {
        self.core_size = Some(size);
        self
    }

    /// Specify the duration for which additional threads outside the
    /// core pool remain alive while not receiving any work before
    /// giving up and terminating.
    pub(crate) fn keep_alive(mut self, keep_alive: Duration) -> Builder {
        self.keep_alive = Some(keep_alive);
        self
    }

    /// Build the `ThreadPool` using the parameters previously supplied
    /// to this `Builder` using the number of CPUs as default core size
    /// if none provided, twice the core size as max size if none
    /// provided, 60 seconds keep_alive if none provided and the default
    /// naming (syd_{pool_number}) if none provided. This
    /// function calls
    /// [`ThreadPool::new`](struct.ThreadPool.html#method.new) or
    /// [`ThreadPool::new_named`](struct.ThreadPool.html#method.new_named)
    /// depending on whether a name was provided.
    pub(crate) fn build(self) -> ThreadPool {
        #[allow(clippy::cast_possible_truncation)]
        let core_size = self
            .core_size
            .unwrap_or_else(|| num_cpus::get().min(usize::from(MAX_SIZE)) as u16);
        let keep_alive = self.keep_alive.unwrap_or_else(|| Duration::from_secs(60));

        ThreadPool::new(core_size, keep_alive)
    }
}

#[derive(Clone)]
struct Worker {
    receiver: crossbeam_channel::Receiver<Job>,
    worker_data: Arc<WorkerData>,
    keep_alive: Option<Duration>,
}

impl Worker {
    fn new(
        receiver: crossbeam_channel::Receiver<Job>,
        worker_data: Arc<WorkerData>,
        keep_alive: Option<Duration>,
    ) -> Self {
        Worker {
            receiver,
            worker_data,
            keep_alive,
        }
    }

    fn try_start(self) -> Result<(), std::io::Error> {
        thread::Builder::new()
            .name("syd_emu".to_string())
            .stack_size(EMU_STACK_SIZE)
            .spawn(move || {
                // Create sentinel.
                let mut sentinel = Sentinel::new(&self);

                // Thread successfully started, increment total and idle
                // worker count, and set CPU affinity.
                let (worker_count, _) = self.worker_data.increment_both();
                // SAFETY: NPROC is never zero.
                #[allow(clippy::arithmetic_side_effects)]
                let core_id = usize::from(worker_count % *NPROC);
                core_affinity::set_for_current(core_affinity::CoreId { id: core_id });

                // Thread is a core thread if keep-alive is None.
                if let Some(keep_alive) = self.keep_alive {
                    while let Ok(task) = self.receiver.recv_timeout(keep_alive) {
                        // Mark current as no longer idle and execute task.
                        self.worker_data.decrement_worker_idle();
                        self.exec_task(&mut sentinel, task);
                    } // If error, either timeout or channel was broken
                      // because sender was disconnected.
                } else {
                    while let Ok(task) = self.receiver.recv() {
                        // Mark current as no longer idle and execute task.
                        self.worker_data.decrement_worker_idle();
                        self.exec_task(&mut sentinel, task);
                    } // If error, channel was broken because sender was
                      // disconnected.
                }

                // Can decrement both at once as the thread only
                // gets here from an idle state (if waiting for work
                // and receiving an error).
                self.worker_data.decrement_both();
            })
            .map(drop)
    }

    #[inline]
    fn exec_task(&self, sentinel: &mut Sentinel, task: Job) {
        sentinel.is_working = true;
        task();
        sentinel.is_working = false;
        // can already mark as idle as this thread will continue the work loop.
        self.worker_data.increment_worker_idle();
    }
}

/// Type that exists to manage worker exit on panic.
///
/// This type is constructed once per `Worker` and implements `Drop` to
/// handle proper worker exit in case the worker panics when executing
/// the current task or anywhere else in its work loop. If the
/// `Sentinel` is dropped at the end of the worker's work loop and the
/// current thread is panicking, handle worker exit the same way as if
/// the task completed normally (if the worker panicked while executing
/// a submitted task) then clone the worker and start it with an initial
/// task of `None`.
struct Sentinel<'s> {
    is_working: bool,
    worker_ref: &'s Worker,
}

impl Sentinel<'_> {
    fn new(worker_ref: &Worker) -> Sentinel<'_> {
        Sentinel {
            is_working: false,
            worker_ref,
        }
    }
}

impl Drop for Sentinel<'_> {
    fn drop(&mut self) {
        if thread::panicking() {
            if self.is_working {
                // Worker thread panicked.
                self.worker_ref.worker_data.decrement_worker_total();
            } else {
                // Idle thread panicked.
                self.worker_ref.worker_data.decrement_both();
            }
        }
    }
}

const WORKER_IDLE_MASK: u32 = 0x0000FFFF;
const INCREMENT_TOTAL: u32 = 0x00010000;
const INCREMENT_IDLE: u32 = 0x00000001;

/// 1. Struct containing data shared between workers.
/// 2. Struct that stores and handles an `AtomicU32` that stores the
///    total worker count in the higher half of bits and the idle worker
///    count in the lower half of bits. This allows to to increment /
///    decrement both counters in a single atomic operation.
#[derive(Default)]
struct WorkerData(AtomicU32);

impl WorkerData {
    #[allow(dead_code)]
    fn increment_both(&self) -> (u16, u16) {
        let old_val = self
            .0
            .fetch_add(INCREMENT_TOTAL | INCREMENT_IDLE, Ordering::Relaxed);
        Self::split(old_val)
    }

    fn decrement_both(&self) -> (u16, u16) {
        let old_val = self
            .0
            .fetch_sub(INCREMENT_TOTAL | INCREMENT_IDLE, Ordering::Relaxed);
        Self::split(old_val)
    }

    fn increment_worker_total(&self) -> u16 {
        let old_val = self.0.fetch_add(INCREMENT_TOTAL, Ordering::Relaxed);
        Self::total(old_val)
    }

    #[allow(dead_code)]
    fn decrement_worker_total(&self) -> u16 {
        let old_val = self.0.fetch_sub(INCREMENT_TOTAL, Ordering::Relaxed);
        Self::total(old_val)
    }

    fn increment_worker_idle(&self) -> u16 {
        let old_val = self.0.fetch_add(INCREMENT_IDLE, Ordering::Relaxed);
        Self::idle(old_val)
    }

    fn decrement_worker_idle(&self) -> u16 {
        let old_val = self.0.fetch_sub(INCREMENT_IDLE, Ordering::Relaxed);
        Self::idle(old_val)
    }

    fn get_total_count(&self) -> u16 {
        Self::total(self.0.load(Ordering::Relaxed))
    }

    fn get_idle_count(&self) -> u16 {
        Self::idle(self.0.load(Ordering::Relaxed))
    }

    #[inline]
    fn split(val: u32) -> (u16, u16) {
        let total_count = (val >> 16) as u16;
        let idle_count = (val & WORKER_IDLE_MASK) as u16;
        (total_count, idle_count)
    }

    #[inline]
    fn total(val: u32) -> u16 {
        (val >> 16) as u16
    }

    #[inline]
    fn idle(val: u32) -> u16 {
        (val & WORKER_IDLE_MASK) as u16
    }
}

struct ChannelData {
    sender: crossbeam_channel::Sender<Job>,
    receiver: crossbeam_channel::Receiver<Job>,
}
